'''
Copyright 2021 Rosalind Franklin Institute

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

#Entry point for doing RL 3D deconvolution using either CPU or GPU

#TODO: A lot of these parameters can be passed as **kwargs.
#TODO: Test GPU acceleration on MacOS
#TODO: Consider open files using scipy.imread rather than the tifffile modules


#from RedLionfishDeconv import helperfunctions
#from helperfunctions import *
#import helperfunctions
import logging
import numpy as np

def doRLDeconvolutionFromNpArrays(data_np , psf_np ,*, niter=10, method='gpu', useBlockAlgorithm=False, callbkTickFunc=None, resAsUint8 = False):
    '''
        Richardson-Lucy deconvolution of 3D data.
        It does NOT use the skimage.image.restoration.rishardson_lucy.
        Iteration uses FFT-based convolutions, either using CPU (scipy) or GPU (Reikna OpenCL)
        
        Parameters:
            data_np: 3d volume of data as numpy array
            psf_np: point-spread-function to use for deconvolution
            niter: number of iterations to perform
            method: 'gpu' to use Reikna OpenCL , 'cpu' to use Scipy. 
            useBlockAlgorithm: 'gpu' only, forces to use block algorithm. Result may show boundary effects.
            callbkTickFunc: function to use to provide tick update during the RL algorithm. This can be either each iteration step or in case of block algorithm, each block calculation
            resAsUint8: Set to return the result of the RL deconvolution as a np.uint8 array. Useful for displaying in napari for example. Note that it will adjust minimum and maximum to ful range 0-255.
                Set to False to return result as default np.float32 format.
    '''
    from RedLionfishDeconv import helperfunctions

    logging.info(f"doRLDeconvolutionFromNpArrays(), niter={niter} , method={method} , useBlockAlgorithm={useBlockAlgorithm}, resAsUint8={resAsUint8}")

    if data_np.ndim !=3:
        logging.info("Data is not 3 dimensional. Exiting.")
        return None
    if psf_np.ndim != 3:
        logging.info("PSF is not 3-dimensional. Exiting.")
        return None

    resRL = None

    if method=='gpu':
        import RedLionfishDeconv.RLDeconv3DReiknaOCL as rlreikna
        if rlreikna.isReiknaAvailable:
            logging.info("Reikna package is available")
            #Great, use the reikna then
            if not useBlockAlgorithm:
                #Use standard RL OCL version
                logging.info("Trying the OCL non-block RL deconvolution.")
                try:
                    resRL = rlreikna.nonBlock_RLDeconvolutionReiknaOCL(data_np, psf_np, niter=niter, callbkTickFunc=callbkTickFunc)
                except Exception as e:
                    #Probably out of memory error, fallback to block algorithm
                    logging.info("nonBlock_RLDeconvolutionReiknaOCL() failed (GPU).")
                    logging.info(e)
                    useBlockAlgorithm= True
            if useBlockAlgorithm:
                logging.info("Trying the OCL block RL deconvolution algorithm.")
                bKeepTrying=True
                blocksize=512

                while bKeepTrying:
                    try:
                        resRL = rlreikna.block_RLDeconv3DReiknaOCL4(data_np , psf_np,niter=niter,max_dim_size=blocksize, callbkTickFunc=callbkTickFunc)
                        bKeepTrying=False
                    except Exception as e:
                        #Error doing previous calculation, reduce block size
                        logging.info(f"Error: block_RLDeconv3DReiknaOCL4 with blocksize={blocksize} failed (GPU). Will try to halve blocksize.")
                        logging.info(e)
                        if blocksize>=128 :
                            blocksize = blocksize//2
                            bKeepTrying=True
                        else:
                            #No point reducing the block size to smaller, fall back to CPU
                            bKeepTrying=False
                            method = 'cpu'
                            logging.info('GPU calculation failed, falling back to CPU.')


        else:
            logging.info("Reikna is not available, falling back to CPU scipy calculation")
            method = 'cpu'
        
    if method == 'cpu':
        from . import RLDeconv3DScipy as rlcpu
        try:
            resRL = rlcpu.doRLDeconvolution_DL2_4(data_np, psf_np, niter=niter, callbkTickFunc=callbkTickFunc)
        except Exception as e:
            logging.info("doRLDeconvolution_DL2_4 failed (CPU) with error:")
            logging.info(str(e))
    
    #TODO: use dask

    if resAsUint8:
        resRL = helperfunctions.convertToUint8AndFullRange(resRL)

    return resRL

def doRLDeconvolutionFromFiles(datapath, psfpath, niter, savepath=None):
    '''
    Opens tiff files and runs Richardson-Lucy deconvoultion
    savepath: location and filename of the tiff file that will store the result

    Returns:
        res_np: result of the RL deconvolution
    '''
    import tifffile as tf

    #Check al info is ok
    data_np = np.array(tf.imread(datapath))
    psf_np = np.array(tf.imread(psfpath))

    res_np = doRLDeconvolutionFromNpArrays(data_np, psf_np, niter=niter, resAsUint8=True)

    if (not res_np is None):
        logging.info("res_np collected")
        if (not savepath is None):
            logging.info(f"Saving data to {savepath}")
            tf.imsave(savepath, res_np)
    
    return res_np




'''
This example shows how to do RL deconvolution using either CPU or GPU,
using callback functionality that gives an update of progress of calculation.
'''

#from RedLionfishDeconv import *
import RedLionfishDeconv as rl
import tifffile

#Load some data
#data = tifffile.imread("test/testdata/gendata_psfconv_poiss_large.tif")
data = tifffile.imread("test/testdata/gendata_psfconv_poiss.tif")
psf = tifffile.imread("test/testdata/PSF_RFI_8bit.tif")

print(f"data.shape: {data.shape}")
print(f"psf.shape: {psf.shape}")

itick=0
def tickCallBackFunction():
    global itick
    print (f"itick: {itick}")
    itick+=1


print("RL Deconvolution: CPU")
res_cpu = rl.doRLDeconvolutionFromNpArrays(data,psf, niter=10, method='cpu',callbkTickFunc=tickCallBackFunction)
#res_cpu = rl.doRLDeconvolutionFromNpArrays(data,psf,niter=10, method='cpu')
print(f"res_cpu.shape: {res_cpu.shape}")
res_cpu_uint8 = rl.helperfunctions.convertToUint8AndFullRange(res_cpu)

itick=0
print("RL Deconvolution: GPU")
#res_gpu = rl.doRLDeconvolutionFromNpArrays(data,psf, 10)
res_gpu = rl.doRLDeconvolutionFromNpArrays(data,psf, niter=10, callbkTickFunc=tickCallBackFunction)
print(f"res_gpu.shape: {res_gpu.shape}")
res_gpu_uint8 = rl.helperfunctions.convertToUint8AndFullRange(res_gpu)

#View result with napari
import napari
np = napari.view_image(res_gpu_uint8, ndisplay=3)
np.add_image(res_cpu_uint8)
napari.run()



'''
Copyright 2021 Rosalind Franklin Institute

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import numpy as np
import logging
import math

#Helper function
def change3DSizeTo(data, shape):
    '''
    Resizes data to the new shape keeping data centralized.
    If data size is smaller, then egdes are cropped.
    If data is larger, data will be padded with zeros.
    '''
    #based in size() code found in
    #https://github.com/Biomedical-Imaging-Group/DeconvolutionLab2/blob/e9af0aba493ba137d70877154648e5583e376a81/src/main/java/signal/RealSignal.java#L531

    mz,my, mx = shape
    nz,ny,nx = data.shape

    psf1 = np.zeros( shape , dtype=data.dtype)
    
    vx = min(nx,mx)
    vy = min(ny,my)
    vz = min(nz, mz)
    ox = int((mx - nx) / 2)
    oy = int((my - ny) / 2)
    oz = int((mz - nz) / 2)

    (pz0, qz0) = (oz, 0) if oz>=0 else (0, -oz)
    (py0, qy0) = (oy, 0) if oy>=0 else (0, -oy)
    (px0, qx0) = (ox, 0) if ox>=0 else (0, -ox)

    psf1[ pz0:pz0+vz , py0:py0+vy , px0:px0+vx ] = data[ qz0:qz0+vz , qy0:qy0+vy ,  qx0:qx0+vx]

    return psf1


#An implementation of the circular() routine from DeconvolutionLab2 that shifts the data centre of the psf
#to the corners of the volume.
#This is important so that convolution of data with psf is will not shift features (like beads).
#It is assumeed that the PSF data has the psf-origin at the center of the data 
def circulify3D(data3D):
    #Check data is 3D
    if data3D.ndim != 3:
        raise Exception("Error, input data is not 3-dimensional")

    data0 = np.array(data3D)

    shape = data0.shape

    zdimhalf0 = int(shape[0]/2)
    zdimhalf1 = shape[0]-zdimhalf0

    ydimhalf0 = int(shape[1]/2)
    ydimhalf1 = shape[1]-ydimhalf0

    xdimhalf0 = int(shape[2]/2)
    xdimhalf1 = shape[2]-xdimhalf0

    if zdimhalf0 >=1 and ydimhalf0>=1 and xdimhalf0>=1:
        data0[zdimhalf1: , ydimhalf1:, xdimhalf1:] = data3D[:zdimhalf0 , :ydimhalf0 , :xdimhalf0 ] #cube 1
        data0[zdimhalf1: , ydimhalf1:, :xdimhalf1] = data3D[:zdimhalf0 , :ydimhalf0 , xdimhalf0: ] #cube 2
        data0[zdimhalf1: , :ydimhalf1, :xdimhalf1] = data3D[:zdimhalf0 , ydimhalf0: , xdimhalf0: ] #cube 3
        data0[zdimhalf1: , :ydimhalf1, xdimhalf1:] = data3D[:zdimhalf0 , ydimhalf0: , :xdimhalf0 ] #cube 4
        data0[:zdimhalf1 , ydimhalf1:, xdimhalf1:] = data3D[zdimhalf0: , :ydimhalf0 , :xdimhalf0 ] #cube 5
        data0[:zdimhalf1 , ydimhalf1:, :xdimhalf1] = data3D[zdimhalf0: , :ydimhalf0 , xdimhalf0: ] #cube 6
        data0[:zdimhalf1 , :ydimhalf1, :xdimhalf1] = data3D[zdimhalf0: , ydimhalf0: , xdimhalf0: ] #cube 7
        data0[:zdimhalf1 , :ydimhalf1, xdimhalf1:] = data3D[zdimhalf0: , ydimhalf0: , :xdimhalf0 ] #cube 8
    
    return data0

def convertToFloat32AndNormalise(data, normaliseType=None, bResetZero=False):
    '''
    Converts data to float32 format and optionally resets zero and normalises
    Parameters:
        data
        normalisetype: None for no normalisation, 'max' normalises to max, 'sum' to normalise to sum of whole volume
        bResetZero: flags to whether shift the minimum value to zero
    '''
    if data is None:
        return None
    
    ret= data.astype(np.float32)

    vmin= 0
    if bResetZero:
        vmin = ret.min()
    ret = ret-vmin

    if not normaliseType is None:
        normcorr = 1.0
        if normaliseType == 'max':
            #Normalise to maximum value
            normcorr= ret.max()
        else:
            #normalise to sum
            normcorr = np.sum(ret)
        #Check for zero before division
        if normcorr==0.0:
            logging.info("normcorr = 0. Normalisation will be skipped to prevent division by zero.")
        else:
            ret = ret/normcorr
    return ret

def convertToUint8AndFullRange(data):
    if data is None:
        return None
        
    res_256 = convertToFloat32AndNormalise(data, normaliseType='max', bResetZero=True)*256
    res_uint8 = res_256.astype('uint8')
    return res_uint8

def generateGaussPSF(shape):
    #shape = (32,32,32)
    z_mg = np.linspace(-shape[0],shape[0] ,shape[0], dtype=np.float32)
    y_mg = np.linspace(-shape[1],shape[1] ,shape[1], dtype=np.float32)
    x_mg = np.linspace(-shape[2],shape[2] ,shape[2], dtype=np.float32)

    z_mg, y_mg, x_mg = np.meshgrid( z_mg, y_mg, x_mg)

    sigma = 5

    data = 1/sigma/math.sqrt(2*math.pi) * np.exp( -0.5 * ( np.square(z_mg) + np.square(y_mg) + np.square(x_mg) ) / sigma/sigma )

    return data


'''
Benchmark RedLionfish deconvolution on a large 3D data.
Generate data here and then run the deconvolution.
Prints the time it takes to complete the deconvolution
'''

import time
import numpy as np
import math
import scipy.signal

import logging
logging.basicConfig(level=logging.INFO)
#logging.info('test logging info')

import RedLionfishDeconv as rl
#If there is an error with the package it will throw an error here
print("import RedLionfishDeconv succeeded")

#Get OpenCL information
print("Checking RLDeconv3DReiknaOCL")
import RedLionfishDeconv.RLDeconv3DReiknaOCL as rlfocl
import RedLionfishDeconv.helperfunctions
rlfocl.printGPUInfo()

testWithGPU = False
if rlfocl.isReiknaAvailable:
    testWithGPU = True
    print("Reikna available. Will run GPU tests.")
else:
    print("Reikna not available. GPU deconvolution tests will be skipped.")
    
#Run a deconvolution

#Create a small 2d array to do FFT
datashape = (256,256,256)
print(f"Creating 3D data with shape {datashape}.")
data = np.zeros(datashape, dtype=np.float32)

#Add a few cubes in grid-like locations
cubesize=2
cubespacing=16
for iz in range(int(cubespacing/2) ,datashape[0],cubespacing):
    for iy in range(int(cubespacing/2) ,datashape[1],cubespacing):
        for ix in range(int(cubespacing/2) ,datashape[2],cubespacing):
            data[iz:iz+cubesize , iy:iy+cubesize , ix:ix+cubesize] = np.ones((cubesize,cubesize,cubesize))

print("Generating PSF data.")

datapsf = RedLionfishDeconv.helperfunctions.generateGaussPSF( (32,32,32) )

print("Convoluting, using scipy.signal")
import scipy.signal
data_convolved = scipy.signal.convolve(data, datapsf, mode='same')
#Normalises it to range 0-255
data_convolved = (data_convolved - data_convolved.min()) / (data_convolved.max() - data_convolved.min())*255

print("Adding poisson noise to data.")
rng = np.random.default_rng()
data_convolved_noised = rng.poisson(lam = data_convolved)

data_convolved_noised_uint8 = ((data_convolved_noised - data_convolved_noised.min()) / ( data_convolved_noised.max() - data_convolved_noised.min() ) *255 ).astype(np.uint8)

niter = 10

print(f"RL deconvolution, niter = {niter}")

import time
print("CPU")

t0 = time.time()
res_CPU = rl.RLDeconvolve.doRLDeconvolutionFromNpArrays(data_convolved_noised_uint8, datapsf, niter=niter, method='cpu')
t1=time.time()

print(f"RL deconvolution using CPU took {t1-t0} s")

if testWithGPU:
    print()
    print("GPU")
    t0 = time.time()
    res_GPU = rl.RLDeconvolve.doRLDeconvolutionFromNpArrays(data_convolved_noised_uint8, datapsf, niter=niter, method='gpu')
    t1=time.time()
    print(f"RL deconvolution using GPU took {t1-t0} s")



print()

#With larger data

#Create a small 2d array to do FFT
datashape = (1024,1024,64)
print(f"Creating large 3D data with shape {datashape}.")
data = np.zeros(datashape, dtype=np.float32)

#Add a few cubes in grid-like locations
cubesize=2
cubespacing=16
for iz in range(int(cubespacing/2) ,datashape[0],cubespacing):
    for iy in range(int(cubespacing/2) ,datashape[1],cubespacing):
        for ix in range(int(cubespacing/2) ,datashape[2],cubespacing):
            data[iz:iz+cubesize , iy:iy+cubesize , ix:ix+cubesize] = np.ones((cubesize,cubesize,cubesize))

print("Convoluting 3D data with the psf, using scipy.signal")
data_convolved = scipy.signal.convolve(data, datapsf, mode='same')
#Normalises it to range 0-255
data_convolved = (data_convolved - data_convolved.min()) / (data_convolved.max() - data_convolved.min())*255

print("Adding poisson noise to data.")
rng = np.random.default_rng()
data_convolved_noised = rng.poisson(lam = data_convolved)

data_convolved_noised_uint8 = ((data_convolved_noised - data_convolved_noised.min()) / ( data_convolved_noised.max() - data_convolved_noised.min() ) *255 ).astype(np.uint8)

print(f"RL deconvolution, niter = {niter}")

print("CPU")

t0 = time.time()
res_CPU = rl.RLDeconvolve.doRLDeconvolutionFromNpArrays(data_convolved_noised_uint8, datapsf, niter=niter, method='cpu')
t1=time.time()

print(f"RL deconvolution using CPU took {t1-t0} s")

if testWithGPU:
    print()
    print("GPU")
    t0 = time.time()
    res_GPU = rl.RLDeconvolve.doRLDeconvolutionFromNpArrays(data_convolved_noised_uint8, datapsf, niter=niter, method='gpu')
    t1=time.time()
    print(f"RL deconvolution using GPU took {t1-t0} s")



'''
Copyright 2021 Rosalind Franklin Institute

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

#Richardson-Lucy using Reikna OpenCL

#Usage will be through a class, whereby user needs to set the shape for the 3D calculation volume
#And then set the psf
#And then run

import logging
#This could be useful to check all ok before trying to run GPU accelerated code
# If not, user can fallback to CPU version using scipy
isReiknaAvailable = False
try:
    import reikna.cluda as cluda
    from reikna.fft import FFT
    from reikna.cluda import functions, dtypes

    apitest = cluda.ocl_api() #Just checking it can get an opencl device (may not be the fastest available!)
    #If it fails to get a device, it will throw an error
    isReiknaAvailable=True
    del(apitest)

except Exception as e:
    logging.error("Failed to setup Reikna with OpenCL.")
    logging.error(e)
    isReiknaAvailable=False

import logging
import numpy as np
from .helperfunctions import *

class RLDeconv3DReiknaOCL:
    '''
    Class that helps setting up and run Richardson Lucy deconvolution using Reikna opencl FFT
    To use this, construct an instance by providing the shape information of the data that you want to do the running the RL algorithm.
    This will setup the calculations in the device, which include compilation steps.
    Before running the RL, user has to set the PSF using function setPSF().
    This will setup the psf internally for making calculations faster,
    such as resizing psf to datahshape, circulify, and precalculate ffts for convolutions.

    Run the RL algorithm by calling doRLDeconvolution

    Note that calculation may fail if data is too large, due to limitations also imposed by PyOpenCL and ReiknaFFT.
    If this is a problem, you should use the block convolution algorithms below.
    '''
    def __init__(self, shape):
        '''
        Parameter
            shape: shape of the data that will be used for the deconvolution algorithm.
            It is important to do it here so that class can do some preparatory steps before running
            the iterative calculation

        '''

        self.shape = shape

        self.api = cluda.ocl_api()
        self.thr = self.api.Thread.create()

        ocldevice = self.api.get_platforms()[0].get_devices()[0] #Get first device available
        logging.info(f"OCL device selected : {ocldevice}")

        devparam = self.api.DeviceParameters(ocldevice)
        self.maxsize = devparam.max_work_item_sizes
        logging.info(f"Device params max_work_item_sizes: {self.maxsize}")

        #check shape is not too large
        if np.product(np.array(shape)) > np.product(np.array(self.maxsize)):
            #logging.error(f"Shape {shape} is too large for OpenCL device shape limits {self.maxsize}")
            # raise ValueError("Shape is too large.")
            logging.info(f"Shape {shape} is too large for OpenCL device shape limits {self.maxsize}. But will try regardless.")

        dtype=np.complex64 #Reikna fft only works with complex types
        #Multiply kernel
        multprogram = self.thr.compile("""
        KERNEL void multiply_3d_cplx_arrays(
            GLOBAL_MEM ${ctype} *a,
            GLOBAL_MEM ${ctype} *b,
            GLOBAL_MEM ${ctype} *dest)
        {
            const SIZE_T id0 = get_global_id(0); 
            const SIZE_T id1 = get_global_id(1);
            const SIZE_T id2 = get_global_id(2);

            int IDX = (id0*get_global_size(1)+id1)*get_global_size(2) +id2;

            //Calculate the product value
            dest[IDX] = ${mul}(a[IDX], b[IDX]); //not sure this multiplication is working correctly

        }
        """, render_kwds=dict(
            ctype=dtypes.ctype(dtype),
            mul=functions.mul(dtype, dtype)))
        #Gets a reference to the kernel
        self.kmultiply_3d_cplx_arrays = multprogram.multiply_3d_cplx_arrays

        #Division kernel a/b
        #Ignores complex part and set complex part to zero
        #Also contains a way to prevent division by zero by setting denominator to 1e-14
        divprogram = self.thr.compile("""
        KERNEL void divide_3d_cplx_arrays_realonly(
            GLOBAL_MEM ${ctype} *a,
            GLOBAL_MEM ${ctype} *b,
            GLOBAL_MEM ${ctype} *dest)
        {
            const SIZE_T id0 = get_global_id(0); 
            const SIZE_T id1 = get_global_id(1);
            const SIZE_T id2 = get_global_id(2);

            int IDX = (id0*get_global_size(1)+id1)*get_global_size(2) +id2;

            float a0 = a[IDX].x;
            float b0= b[IDX].x;

            if (b0 ==  0.0) {
                b0= 1e-14;
            };

            //Calculate the division value
            //dest[IDX] = ${div}(a[IDX], b[IDX]);
            //dest[IDX] = ${div}(a[IDX], b0);
            dest[IDX].x = a0/b0;
            dest[IDX].y = 0;

        }
        """, render_kwds=dict(
            ctype=dtypes.ctype(dtype),
            div=functions.div(dtype, dtype)))
        self.kdivide_3d_cplx_arrays_realonly = divprogram.divide_3d_cplx_arrays_realonly

        #Prepare psf fft data in device
        self.psf_fft_dev = self.thr.array(shape, dtype)
        self.psf_flip_fft_dev = self.thr.array(shape, dtype)
        self.is_psf_set = False

        #Prepare the fft kernel using the psf_fft_dev as reference
        self.cfft_gpu = FFT(self.psf_fft_dev).compile(self.thr)


    def setPSF(self,psfdata):
        #Prepare data
        #Convert to float, normalise to sum, calculate fft's and of flipped version
        logging.info("setPSF()")

        psf_norm = convertToFloat32AndNormalise(psfdata , normaliseType='sum', bResetZero=False) #Normalise to sum
        psf0 = change3DSizeTo(psf_norm, self.shape) #Adjust size
        psf1 = circulify3D(psf0) #circulify psf
        psf_cplx = psf1.astype(np.complex64)

        #precalculate psf ffts
        psf_dev = self.thr.to_device(psf_cplx)
        self.cfft_gpu(self.psf_fft_dev , psf_dev) #fft calculation, stores in self.psf_fft_dev
        del(psf1)
        del(psf_dev) #Clear device memory
        #WARNING: If psf is too large, it will throw an error that cannot be caught.
        #

        psf_cplx_flip = np.array(np.flip(psf_cplx))
        psf_flip_dev = self.thr.to_device(psf_cplx_flip)
        #psf_flip_fft_dev = thr.array(shape, np.complex64)
        self.cfft_gpu(self.psf_flip_fft_dev , psf_flip_dev) #fft calculation, stores in self.psf_flip_fft_dev
        del(psf_cplx_flip)
        del(psf_flip_dev)

        self.is_psf_set = True
    
    def doRLDeconvolution(self, data_np, *, niter = 10, callbkTickFunc = None):
        '''
        Does the Richardson Lucy Deconvolution using Reikna OpenCL
        with custom mulitply and division kernels
        
        returns None if fail
        returns the Deconvoluted data volume, only the real part (float32 format)
        '''
        
        logging.info("doRLDeconvolution()")

        #Check data and shape are consistent
        if data_np.shape[0] != self.shape[0] or data_np.shape[1] != self.shape[1] or data_np.shape[2] != self.shape[2]:
            logging.error("Data input shape is different from the shape that was initially set. Please make sure shapes are the same.")
            return None
        
        #Check psf was set
        if self.is_psf_set:

            #Check data. data must be float type
            #No need to check because it will converted to complex anyway
            if not data_np.dtype is np.dtype(np.float32):
                print ("data_np is not np.float32 type. Exiting")
                return None

            #data0 = data_np
            data_cplx = data_np.astype(np.complex64)
            #Send data to device
            data_cplx_dev = self.thr.to_device(data_cplx)

            #ready for the RL iterations

            #Prepare tempary storage on device
            xn1 =  np.array(data_cplx) #initialise copy
            xn1_dev = self.thr.to_device(xn1)
            xn_buff = self.thr.empty_like(xn1_dev)

            t0_dev = self.thr.empty_like(xn1_dev)
            t1_dev = self.thr.empty_like(xn1_dev)

            for i in range(niter):
                xn_dev = xn1_dev

                #Convolution psf*xn
                self.cfft_gpu(t0_dev, xn_dev) # FFT of xn
                self.kmultiply_3d_cplx_arrays( t0_dev , self.psf_fft_dev, t1_dev , local_size=(1,1,1), global_size=self.shape )
                self.cfft_gpu(t0_dev, t1_dev, 1) # FFT inverse, result to t0_dev

                #Division:  image / (psf*xn) , result in t1_dev
                self.kdivide_3d_cplx_arrays_realonly( data_cplx_dev , t0_dev , t1_dev , local_size=(1,1,1), global_size=self.shape )

                #Second convolution with psf_flipped
                self.cfft_gpu(t0_dev, t1_dev) # FFT of t1_dev, result to t0_dev
                self.kmultiply_3d_cplx_arrays( self.psf_flip_fft_dev , t0_dev, t1_dev , local_size=(1,1,1), global_size=self.shape ) #multiplication, result to t1_dev
                self.cfft_gpu(t0_dev, t1_dev, 1) # FFT inverse, result in t0_dev

                #Multiply with xn
                self.kmultiply_3d_cplx_arrays( xn_dev , t0_dev, xn_buff , local_size=(1,1,1), global_size=self.shape ) #multiplication, result to xn_buff

                #Swap buffers
                xn1_dev = xn_buff
                xn_buff = xn_dev

                if not callbkTickFunc is None:
                    callbkTickFunc()

            #Collect result
            xn1 = xn1_dev.get().real
            logging.info("RL completed, result collected")
            
            logging.info("Clearing GPU RAM")
            #Clear GPU memory
            del(xn1_dev)
            del(xn_buff)
            del(t0_dev)
            del(t1_dev)
            #del(psf_flip_fft_dev)
            #del(psf_fft_dev)
            del(data_cplx_dev)

            return xn1

        else:
            logging.error("Psf was not set. Please set it by doing .setPSF(psfdata) .")
            return None
    
    @staticmethod
    def getDefaultDeviceMaxSize():
        api = cluda.ocl_api()

        ocldevice = api.get_platforms()[0].get_devices()[0] #Get first device available
        devparam = api.DeviceParameters(ocldevice)
        maxsize = devparam.max_work_item_sizes
        del(api)
        return maxsize

def nonBlock_RLDeconvolutionReiknaOCL( data_np, psf_np, *, niter = 10, callbkTickFunc=None):
    '''
    This will do the RL deconvolution from 3D data_np using the psf_np provided.
    This does not use block iteration so large arrays may throw out of memory errors

    '''
    logging.info("nonBlock_RLDeconvolutionReiknaOCL()")
    
    if data_np.ndim !=3 or psf_np.ndim!=3:
        logging.error ("Data and psf data must be 3 dimensional. Exiting.")
        return None
    
    data = convertToFloat32AndNormalise(data_np,None,bResetZero=False)
    
    shape = data_np.shape
    myRL_Reikna =  RLDeconv3DReiknaOCL(shape) #Set up calculation by instantiating the class RLDeconv3DReiknaOCL
    myRL_Reikna.setPSF(psf_np)
    
    return myRL_Reikna.doRLDeconvolution(data, niter=niter, callbkTickFunc=callbkTickFunc)

def _isShapeTooBigForDevice(shape):
    ret=False
    
    maxsize = RLDeconv3DReiknaOCL.getDefaultDeviceMaxSize()

    mult_maxsize= np.product(np.array(maxsize))
    mult_shape = np.product(np.array(shape))

    if mult_shape>mult_maxsize:
        ret=True

    return ret


#Default
def block_RLDeconv3DReiknaOCL4(data, psfdata, *, niter=10, max_dim_size=256, psfpaddingfract = 1.2, callbkTickFunc=None):
    '''
    In this version, blockstep is reduced, effectively setting the valid area to a smaller part of the block calculation.
    New parameter psfpaddingfract to set how how much padding relative to psfsize to use

    Parameters
        data: numpy 3D array with the data to deconvolve
        psfdata: numpy 3D array with the PSF to use for the deconvolution
        niter: number of Richardson-Lucy algorithm iterations
        paddingfract: padding to use when merging data as a relative fraction on psf size
    '''
    logging.info("block_RLDeconv3DReiknaOCL4()")

    if data.ndim !=3 or psfdata.ndim!=3:
        logging.error("Data and psf data must be 3 dimensional. Exiting.")
        raise ValueError("Data or psf are not 3D")
    
    #Check psf size is not too large for the GPU calculation
    if _isShapeTooBigForDevice(psfdata.shape):
        logging.WARNING("PSF data shape is larger than GPU data size limits. Error may occur")
        #raise ValueError("Psf data shape is too large")

    data = convertToFloat32AndNormalise(data,bResetZero=False)

    shapedata = data.shape
    shapepsf = psfdata.shape

    logging.info(f"data shape: {shapedata} , psf shape: {shapepsf}")

    #For each of the dimensions.
    #check size is larger than max_dim_size. If it is then limit block calculation using max_dim_size
    blockshape=[0,0,0]
    for axis in range(3):
        blockshape0 = max_dim_size #default
        if shapedata[axis]<max_dim_size :
            blockshape0 = shapedata[axis]
        blockshape[axis] = blockshape0
    
    logging.info(f"blockshape: {blockshape}")

    # Check for each axis whether PSF is larger and if blocking will be done along that axis
    # If yes, then don't use this algorithm as it will give bad results
    for axis in range(3):
        if shapedata[axis] > blockshape[axis] and shapepsf[axis]> blockshape[axis]:
            raise ValueError("PSF and data does not allow efficient calculation using this blocking algorithm. Exiting.")

    #check if shape is too large
    if _isShapeTooBigForDevice(blockshape):
        #logging.error("blockshape is too large. Exiting.")
        logging.WARNING("Adjusted blockshape is larger than GPU data size limits. Error may occur")
        #raise ValueError("blockshape is too large")

    #Set step (and padding) from the blockshape and padding being psf size *1.5
    #Valid area being the block size minus the psf size * psfpaddingfract

    #blockstep = list(shapedata) #default, makes a copy
    #validshape = list(shapedata)

    blockstep = list(blockshape) #default, makes a copy
    #validshape = list(blockshape)
    for a in range(3):
        v0 = int(blockshape[a] - psfpaddingfract*shapepsf[a])
        if blockshape[a]< shapedata[a] and v0>0:
            #validshape[a] = v0
            #blockstep[a] =  validshape[a]
            blockstep[a]=v0

    validshape = blockstep
    logging.info(f"blockstep: {blockstep}")

    datares = np.zeros(data.shape) #To collect results

    #Setup Reikna RL Deconv Class to use this shape
    my_rldeconv = RLDeconv3DReiknaOCL(blockshape)
    my_rldeconv.setPSF(psfdata)

    #do the block iteration, for loops for each dimension
    # the values with zero of the loop are for the left corner
    #The indexes i are for original data
    for iz0 in range(0,shapedata[0], blockstep[0]):
        iz00=iz0
        iz1 = iz0 + blockshape[0]
        if iz1>shapedata[0]:
            iz1 = shapedata[0]
            iz00 = iz1 - blockshape[0]
            if iz00<0: iz00=0
        
        for iy0 in range(0,shapedata[1], blockstep[1]):
            iy00 = iy0
            iy1 = iy0 + blockshape[1]
            if iy1>shapedata[1]:
                iy1 = shapedata[1]
                iy00 = iy1 - blockshape[1]
                if iy00<0: iy00=0

            for ix0 in range(0,shapedata[2], blockstep[2]):
                ix00 = ix0
                ix1 = ix0 + blockshape[2]
                if ix1>shapedata[2]:
                    ix1 = shapedata[2]
                    ix00 = ix1 - blockshape[2]
                    if ix00<0: ix00=0

                logging.info(f"New block, intended origin iz0,iy0,ix0 = {iz0},{iy0},{ix0} , use origin iz00,iy00,ix00 = {iz00},{iy00},{ix00} , end iz1,iy1,ix1 = {iz1},{iy1},{ix1}")

                #Get the data block
                datablock0 = data[iz00:iz1, iy00:iy1, ix00:ix1]
                
                logging.info("Start RL deconvolution of this block")
                #Do RL with this datablock
                rl_of_datablock = my_rldeconv.doRLDeconvolution(datablock0,niter=niter)
                logging.info("This block's RL deconvolution completed")
                
                #Store the datablock result, only the valid part
                #unless it is the leftmost (first block) of the dimension given
                jz0=0
                jy0=0
                jx0=0

                #Crop the padded on the left side
                if iz0 !=0 :
                    jz0 += int( (blockshape[0] - validshape[0]) / 2)
                if iy0 !=0:
                    jy0 += int( (blockshape[1] - validshape[1]) / 2)
                if ix0 !=0:
                    jx0 += int( (blockshape[2] - validshape[2]) / 2)
                
                logging.info(f"Crop block result from origin jz0,jy0,jx0 = : {jz0},{jy0},{jx0}")
                
                logging.info(f"Copying cropped block to datares")
                datares[ iz00+jz0 : iz00+rl_of_datablock.shape[0] , iy00+jy0 : iy00+rl_of_datablock.shape[1] , ix00+jx0 : ix00+rl_of_datablock.shape[2]] = rl_of_datablock[jz0: , jy0: , jx0: ]

                if not callbkTickFunc is None:
                    callbkTickFunc()

    logging.info("Completed block RL deconvolution.")

    #Hopefully, at this point, datares should have the final result of the blocked RL deconvolution
    return datares
    

def printGPUInfo():
    if isReiknaAvailable:
        api = cluda.ocl_api()

        #code from test_reiknaFFT\reikna_getInfo.ipynb
        devices = api.get_platforms()[0].get_devices()
        print(f"OpenCL devices: {devices}")

        for i in range(len(devices)):
            ocldevice = devices[i]
            print(f"Device {i}: {ocldevice}")
            print(f"max_work_group_size: {api.DeviceParameters(ocldevice).max_work_group_size}") # max_work_group_size: 1024
            print(f"max_work_item_sizes: {api.DeviceParameters(ocldevice).max_work_item_sizes}") # max_work_item_sizes: [1024, 1024, 64]
            print(f"max_num_groups: {api.DeviceParameters(ocldevice).max_num_groups}") # max_num_groups: [18446744073709551616, 18446744073709551616, 18446744073709551616]
            print(f"local_mem_size: {api.DeviceParameters(ocldevice).local_mem_size}") # local_mem_size: 49152
            print(f"local_mem_banks: {api.DeviceParameters(ocldevice).local_mem_banks}") # local_mem_banks: 32

        del(api)
    else:
        print("No Reikna.")




'''
Copyright 2021 Rosalind Franklin Institute

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

#RedLionfish napari plugin

'''
Some help designing this collected from diffrent places

https://napari.org/plugins/stable/hook_specifications.html#gui-hooks
https://github.com/DragaDoncila/example-plugin/blob/main/example_plugin/_dock_widget.py
These examples show adding widget using decorator @magic_factory.
This decorator is poorly documented. I don't know how to use it.


'''

from magicgui import magic_factory
from napari_plugin_engine import napari_hook_implementation

#As suggested by DragaDoncila
# https://github.com/rosalindfranklininstitute/RedLionfish/pull/5#issuecomment-1239989472
# If napari is not installed
# from typing import TYPE_CHECKING
# print(f"TYPE_CHECKING: {TYPE_CHECKING}")
# if TYPE_CHECKING:
#     from napari.types import LabelsData, ImageData, LayerDataTuple
#     from napari.layers import Image

from napari.types import ImageData, LayerDataTuple
from napari.utils import progress

import RedLionfishDeconv as rl

# import logging
# logging.basicConfig(level=logging.INFO)

#Logging in napari is not working not sure why.
# Maybe napari also uses logging and outputs to different location

# To output to console use print or loguru instead.

import time


''' The parameters after the decorator setup the title and other properties of the widget window.

The parameters in the FUNCTION will set the elements that will be shown, but things like comboboxes
or sliders will have no limits set, this should be done in the decorator.

see also
https://napari.org/guides/stable/magicgui.html

Note that the @magic_factory and @magicgui behave in similar way.
'''

#The widget
@magic_factory (
    call_button="Go" ,
    iterations={'max':16384},
    useGPU={"label": "Use GPU if possible"}, # https://github.com/napari/magicgui/blob/main/examples/change_label.py
    resAsUInt8={"label": "Result as clip-norm integer (uint8)"}
    )
def RedLionfish_widget(
    data: ImageData, #Input is data that can be selected
    psfdata:ImageData,
    iterations=10, #User chooses number of RL iterations
    useGPU = True,
    resAsUInt8=True
    ) -> LayerDataTuple: #Result is a LayerDataTuple, like (data, {dict_properties})

    print(f"iterations = {iterations}")

    ret = None

    datares=None
    if not data is None and not psfdata is None:
        #Print information before calculation
        print(f"data.shape = {data.shape} , type(data) = {type(data)} , data.dtype = {data.dtype}")
        print(f"psfdata.shape = {psfdata.shape} , type(psfdata) = {type(psfdata)} , psfdata.dtype = {psfdata.dtype}")

        pbr=progress(total=iterations)
        #pbr=progress()
        pbr.set_description(f"Calculation progress")

        def callback():
            #print("iteration tick")
            pbr.update(1)
            pbr.refresh()
            #time.sleep(0.1)

        if data.ndim==3 and psfdata.ndim==3: #For now, only 3D is supported
            #logging.basicConfig(level=logging.INFO) #not working

            method = 'cpu' #Default
            if useGPU:
                method='gpu'
            #Run the deconvolution
            datares = rl.doRLDeconvolutionFromNpArrays(data, psfdata, niter= iterations, method = method ,resAsUint8=resAsUInt8 , callbkTickFunc=callback)

            ret = ( datares , { 'name':'RL-deconvolution'})
        else:
            print("Data or PSF is not 3-dimensional.")

        pbr.close()
    return ret



@napari_hook_implementation
def napari_experimental_provide_dock_widget():
    # you can return either a single widget, or a sequence of widgets
    #return [test_widget0, test_widget1]
    #return [RedLionfish_widget,test_widget2, test_widget1]
    return RedLionfish_widget


#Testing in local computer using VScode, without having installed RedLionfish package
#import sys
#sys.path.append('..')

#from RedLionfishDeconv import *
import RedLionfishDeconv as rl
import tifffile

#Load some data
data = tifffile.imread("test/testdata/gendata_psfconv_poiss_large.tif")
#data = tifffile.imread("test/testdata/gendata_psfconv_poiss.tif")
psf = tifffile.imread("test/testdata/PSF_RFI_8bit.tif")

print(f"data.shape: {data.shape}")
print(f"psf.shape: {psf.shape}")

res_gpu = rl.doRLDeconvolutionFromNpArrays(data,psf, niter=10,resAsUint8=True)
print(f"res_gpu.shape: {res_gpu.shape}")
#res_gpu_uint8 = rl.helperfunctions.convertToUint8AndFullRange(res_gpu)


import napari
np = napari.view_image(res_gpu, ndisplay=3)
napari.run()



'''
Copyright 2021 Rosalind Franklin Institute

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

#Several versions of Richardson-Lucy deconvolution using scipy.signal convolution

import numpy as np
import scipy.fft

from .helperfunctions import *
import logging

def doRLDeconvolution12(data_np , psf_np , *, niter=10, callbkTickFunc=None):
    #RL deconvolution based on doRLDeconvolution11()
    # Reverse engineered parts of https://scikit-image.org/docs/dev/api/skimage.restoration.html#skimage.restoration.richardson_lucy
    # and also scipy.signal convolution fft code for faster processing
    #https://github.com/scipy/scipy/blob/v1.7.1/scipy/signal/signaltools.py#L1293-L1413
    #for mode='same', method='fft', fftconvolution() #
    #There is a small difference here compared to scikit-image version in the sense that the psf is normalised to the sum.
    #This version tries to do all the work with arrays with preferred size of next_fast_len
    #Only at the end of the calculation it gets the shape needed

    #Convert and normalise
    data_np_norm = convertToFloat32AndNormalise(data_np, bResetZero=False)
    psf_np_norm = convertToFloat32AndNormalise(psf_np , normaliseType='sum', bResetZero=False) #Normalise to sum

    s1 = data_np_norm.shape
    s2 = psf_np_norm.shape

    shape = [(s1[i] + s2[i] - 1) for i in range(data_np_norm.ndim)]
    fshape = [scipy.fft.next_fast_len(shape[a], True) for a in range(len(shape))]

    #Resize both the data and psf to the preferred fshape
    data_shaped = change3DSizeTo(data_np_norm, fshape)
    psf_shaped = change3DSizeTo(psf_np_norm, fshape)

    #Circulify psf
    psf1 = circulify3D(psf_shaped)

    #Precalculated psf_fft
    psf_fft= scipy.fft.rfftn(psf1)
    psf_flip = np.flip(psf1)
    psf_flip_fft = scipy.fft.rfftn(psf_flip) #FFT of the flipped psf

    xn1 = np.array(data_shaped) #initialize copy
    for i in range(niter):
        xn=xn1
        xn_fft = scipy.fft.rfftn(xn) 
        U0 = np.multiply(xn_fft ,psf_fft)
        u0 = scipy.fft.irfftn(U0)

        u0 = np.where(u0==0, 1e-14, u0) #Fix zero values giving a finite value
            
        p = np.divide(data_shaped,u0)

        U1 = scipy.fft.rfftn(p)
        U2 = np.multiply(U1 , psf_flip_fft)
        u2 = scipy.fft.irfftn(U2)

        xn1 = np.multiply(xn, u2)

        if not callbkTickFunc is None:
            callbkTickFunc()
    
    #Get the central part of the result
    #xn1_central = _centered(xn1, s1)
    xn1_central = change3DSizeTo(xn1, s1) #To original shape

    # data_deconv_norm_256 = convertAndNormalise(xn1_central)*256
    # data_deconv_uint8 = data_deconv_norm_256.astype('uint8')
    
    data_deconv_uint8 = convertToUint8AndFullRange(xn1_central)
    return data_deconv_uint8


#default
def doRLDeconvolution_DL2_4(data_np , psf_np ,*, niter=10, callbkTickFunc=None):
    #RL deconvolution based in DeconvolutionLab2 algorithm
    #Mimics DeconvolutionLab2 (DL2) as best as possible
    #https://github.com/Biomedical-Imaging-Group/DeconvolutionLab2/blob/master/src/main/java/deconvolution/algorithm/RichardsonLucy.java
    #However it prepares the psf to have the same size as the input data
    #In DeconvolutionLab2 only psf is normalised to psf sum
    #
    #This version uses scipy fft
    #It is modified from the DeconvolutionLab2 algorithm
    # the fft of the psf is not conjugated, but instead is flipped before conjugation
    # The developers probably thought that the 'reality condition' applies
    # FT[f](-y) = FT[f]*
    # but that is not true, as code tests proved (tests.ipynb)
    # In this version, as in the original formula for RL, the flipped psf is used for the second convolution
    # (see tests.ipynb)
    
    logging.info("doRLDeconvolution_DL2_4() (CPU)")
    #Convert and normalise
    data_np_norm =convertToFloat32AndNormalise(data_np, bResetZero=False) #don't normalise

    #Check last axis is even size, otherwise it will give error
    if data_np_norm.shape[-1] %2 != 0: #odd number
        logging.info("Data last axis size is an odd number. Padding with zeros to make size even to prevent errors")
        s = data_np_norm.shape
        shape_fix = (s[0], s[1], s[2]+1)
        data_fix = np.empty(shape_fix, dtype = data_np_norm.dtype)
        data_fix[:,:,:-1] = data_np_norm[:,:,:]
        data_fix[:,:,-1] = 0
        data_np_norm = data_fix

    # psf_norm = psf0/sum#
    psf_norm = convertToFloat32AndNormalise(psf_np, normaliseType='sum',bResetZero=False)

    psf0 = change3DSizeTo(psf_norm, data_np_norm.shape)
    #psf1 should have the same dimensions as data to successfully multiply and divide ffts

    #Do circularisation of psf data before starting the RL algorithm
    psf1 = circulify3D(psf0)

    #Precalculated psf_fft
    psf_fft= scipy.fft.rfftn(psf1)
    psf_flip = np.flip(psf1)
    psf_flip_fft = scipy.fft.rfftn(psf_flip) #FFT of the flipped psf

    xn1 = np.array(data_np_norm) #initialize copy
    for i in range(niter):
        xn=xn1
        xn_fft = scipy.fft.rfftn(xn)  
        U0 = np.multiply(xn_fft ,psf_fft)
        u0 = scipy.fft.irfftn(U0)

        u0 = np.where(u0==0, 1e-14, u0) #Fix zero values giving a finite value

        p = np.divide(data_np_norm,u0)

        U1 = scipy.fft.rfftn(p)
        U2 = np.multiply(U1 , psf_flip_fft)
        u2 = scipy.fft.irfftn(U2)

        xn1 = np.multiply(xn, u2)

        if not callbkTickFunc is None:
            callbkTickFunc()

    dataret = xn1
    return dataret


'''
Benchmark RedLionfish deconvolution on a large 3D data.
Generate data here and then run the deconvolution.
Prints the time it takes to complete the deconvolution
'''

import time
import numpy as np
import scipy.signal

import logging
logging.basicConfig(level=logging.INFO)
#logging.info('test logging info')

import RedLionfishDeconv as rl
#If there is an error with the package it will throw an error here
print("import RedLionfishDeconv succeeded")

#Get OpenCL information
print("Checking RLDeconv3DReiknaOCL")
import RedLionfishDeconv.RLDeconv3DReiknaOCL as rlfocl
import RedLionfishDeconv.helperfunctions
rlfocl.printGPUInfo()

niter = 5

print("Testing weird shape and data.shape = psf.shape")
sameshape = (48,57,93)

print(f"Generating data with shape ={sameshape}")

data = np.zeros(sameshape, dtype=np.float32)
#Add a few cubes in grid-like locations
cubesize=2
cubespacing=16
for iz in range(int(cubespacing/2) ,sameshape[0],cubespacing):
    for iy in range(int(cubespacing/2) ,sameshape[1],cubespacing):
        for ix in range(int(cubespacing/2) ,sameshape[2],cubespacing):
            try:
                data[iz:iz+cubesize , iy:iy+cubesize , ix:ix+cubesize] = np.ones((cubesize,cubesize,cubesize))
            except Exception as e:
                pass #do not put the small cube in case of error

print(f"Generating psf with shape {sameshape}")

datapsf = RedLionfishDeconv.helperfunctions.generateGaussPSF( sameshape )

print("Convoluting 3D data with the psf, using scipy.signal")
data_convolved = scipy.signal.convolve(data, datapsf, mode='same')
#Normalises it to range 0-255
data_convolved = (data_convolved - data_convolved.min()) / (data_convolved.max() - data_convolved.min())*255

print("Adding poisson noise to data.")
rng = np.random.default_rng()
data_convolved_noised = rng.poisson(lam = data_convolved)

data_convolved_noised_uint8 = ((data_convolved_noised - data_convolved_noised.min()) / ( data_convolved_noised.max() - data_convolved_noised.min() ) *255 ).astype(np.uint8)

print(f"RL deconvolution, niter = {niter}")
print("CPU")
t0 = time.time()
res_CPU = rl.RLDeconvolve.doRLDeconvolutionFromNpArrays(data_convolved_noised_uint8, datapsf, niter=10, method='cpu', resAsUint8=True)
t1=time.time()
print(f"RL deconvolution using CPU took {t1-t0} s")

print("GPU")
t0 = time.time()
res_GPU = rl.RLDeconvolve.doRLDeconvolutionFromNpArrays(data_convolved_noised_uint8, datapsf, niter=10, method='gpu', resAsUint8=True)
t1=time.time()
print(f"RL deconvolution using GPU took {t1-t0} s")

datapsf_uint8 = RedLionfishDeconv.helperfunctions.convertToUint8AndFullRange(datapsf)

print("skimage.restoration.richardson_lucy")
#Compare with skimage
import skimage.restoration
t0 = time.time()
res_skimage = skimage.restoration.richardson_lucy(data, datapsf, iterations=10,clip=False)
t1=time.time()
print(f"RL deconvolution using scikitimage took {t1-t0} s")

import napari
nv = napari.view_image(res_GPU, ndisplay=3)
nv.add_image(res_CPU)
nv.add_image(data_convolved_noised_uint8)

nv.add_image(datapsf_uint8)

nv.add_image(res_skimage)

napari.run()




#from RedLionfishDeconv import *
import RedLionfishDeconv as rl
import tifffile

import logging
logging.basicConfig(level=logging.INFO)

#Load some data
#data = tifffile.imread("test/testdata/gendata_psfconv_poiss_large.tif")
data = tifffile.imread("test/testdata/gendata_psfconv_poiss.tif")
psf = tifffile.imread("test/testdata/PSF_RFI_8bit.tif")

print(f"data.shape: {data.shape}")
print(f"psf.shape: {psf.shape}")

#test
res_cpu = rl.doRLDeconvolutionFromNpArrays(data,psf, niter=10, method='cpu')
res_cpu_uint8 = rl.helperfunctions.convertToUint8AndFullRange(res_cpu)


res_gpu = rl.doRLDeconvolutionFromNpArrays(data,psf, niter=10)
print(f"res_gpu.shape: {res_gpu.shape}")
res_gpu_uint8 = rl.helperfunctions.convertToUint8AndFullRange(res_gpu)

print("skimage.restoration.richardson_lucy")
#Compare with skimage
import skimage.restoration
res_skimage = skimage.restoration.richardson_lucy(data, psf, iterations=10,clip=False)
print("Completed skimage.restoration.richardson_lucy.")
print(f"res_sk.shape: {res_skimage.shape}")
res_skimage_uint8 = rl.helperfunctions.convertToUint8AndFullRange(res_skimage)

import napari
nv = napari.view_image(data, ndisplay=3)
nv.add_image(res_skimage_uint8)
nv.add_image(res_gpu_uint8)
nv.add_image(res_cpu_uint8)

napari.run()



'''
Copyright 2021 Rosalind Franklin Institute

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from setuptools import setup

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    version = '0.7',
    name = 'RedLionfish',
    description = 'Fast Richardson-Lucy deconvolution of 3D volume data using GPU or CPU with napari plugin.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url = 'https://github.com/rosalindfranklininstitute/RedLionfish',
    author = 'Luis Perdigao',
    author_email='luis.perdigao@rfi.ac.uk',
    packages=['RedLionfishDeconv'],
    classifiers=[
        'Development Status :: 4 - Beta ',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Image Processing',
        'Framework :: napari'
    ],
    license='Apache License, Version 2.0',
    zip_safe=False,
    install_requires=[
        'numpy',
        'scipy',
        'pyopencl',
        'reikna'
    ],

    #for the napari plugin
    entry_points={'napari.plugin': 'RedLionfish = RedLionfishDeconv.napari_plugin'}
)
